#### Packages
library(LaplacesDemon)
library(lsa)
library(igraph)
library(aricode)
library(pdfCluster)
library(hydra)
library(igraph)
library(xtable)

#### Main function
cs_sbm <- function(n_cluster, 
                   net_adj,
                   alpha = rep(2, n_cluster), 
                   Z.initial = matrix(rep(1/n_cluster, n_cluster * dim(net_adj)[1]), nrow = n_cluster), 
                   burnin = 5000, 
                   size  = 10000,
                   thin = 5){
  iter <- 0
  
  n <- dim(net_adj)[1]
  
  Z <- Z.initial
  
  Z.res <- array(NA, dim = c(n_cluster, n, size))
  
  while(iter < burnin + size){
    
    for(i in 1:n){
      
      Z.prop <- as.vector(LaplacesDemon::rdirichlet(n = 1, alpha = alpha))
      
      cs.prop <- sapply(c(1:n), function(x){lsa::cosine(Z.prop, Z[,x])})
      
      num.prop <- prod(cs.prop[-i]^net_adj[i, -i]*(1 - cs.prop[-i])^(1 - net_adj[i, -i]))*prod(Z.prop^(alpha - 1))
      
      cs.old <- sapply(c(1:n), function(x){lsa::cosine(Z[,i], Z[,x])})
      
      den.prop <- prod(cs.old[-i]^net_adj[i, -i]*(1 - cs.old[-i])^(1 - net_adj[i, -i]))*prod(Z[,i]^(alpha - 1))
      
      accept.rate <- num.prop/den.prop
      
      U <- runif(1, 0, 1)
      
      if (U < accept.rate){
        Z[,i] <- Z.prop
      } else{
        Z[,i] <- Z[,i]
      }
    }
    
    iter = iter + 1
    
    if(iter > burnin){
     
      Z.res[,,iter - burnin] <- Z 
    }
  }
  
  Z.posterior <- Z.res[,,seq(1, size, thin)]
  
  Z.output <- as.data.frame(apply(Z.posterior, 1:2, mean))
  
  rownames(Z.output) <- paste0("cluster", 1:n_cluster)
  
  colnames(Z.output) <- paste0("V", 1:n)
  
  return(list("mixed.member" = Z.output,
              "clustering" = apply(Z.output, 2, which.max)
              )
         )
}

#### Zachary karate club
set.seed(314159)
data(karate)
karate.adj <- karate$adjacency
karate.res <- cs_sbm(n_cluster = 2, net_adj = karate.adj, burnin = 5000, size = 10000)
saveRDS(karate.res, "karate_res.rds")
karate.graph <- igraph::graph_from_adjacency_matrix(karate.adj, mode = "undirected")
karate.layout <- layout.fruchterman.reingold(karate.graph)
V(karate.graph)$group <- karate.res$clustering

pdf(file = file.path("..", "figures", "karate_plot.pdf"))
par(mar = c(0.5, 0.5, 0.1, 0.1))
plot(karate.graph, 
     layout = karate.layout, 
     vertex.color = V(karate.graph)$group, 
     vertex.size = 13)
dev.off()

pdf(file = file.path("..", "figures", "karate_truth.pdf"))
par(mar = c(0.5, 0.5, 0.1, 0.1))
V(karate.graph)$group <- karate$group
plot(karate.graph, 
     layout = karate.layout, 
     vertex.color = V(karate.graph)$group, 
     vertex.size = 13)
dev.off()

karate.df <- data.frame(t(karate.res$mixed.member[1,]), 
                        t(karate.res$mixed.member[2,]),
                        karate.res$clustering)
xtable(karate.df, digits = c(0, 4, 4, 0))

#### Bottlenose dolphin
dolphins.graph <- read.graph(file='https://networkdata.ics.uci.edu/data/dolphins/dolphins.gml', 
                             format='gml')
dolphins.adj <- as.matrix(igraph::as_adjacency_matrix(dolphins.graph))
dolphins.res <- cs_sbm(n_cluster = 2, net_adj = dolphins.adj, burnin = 50000, size = 50000)
saveRDS(dolphins.res, "dolphins_res.rds")
dolphins.layout <- layout.fruchterman.reingold(dolphins.graph)
V(dolphins.graph)$group <- dolphins.res$clustering

pdf(file = file.path("..", "figures", "dolphins_plot.pdf"))
par(mar = c(0.5, 0.5, 0.1, 0.1))
plot(dolphins.graph, 
     layout = dolphins.layout, 
     vertex.color = V(dolphins.graph)$group, 
     vertex.size = 13)
dev.off()

dolphins.truth <- c("Beescratch", "DN16", "DN21", "DN63", "Feather", "Gallatin", "Jet", "Knit", 
                 "Mus", "MN23","Notch", "Number1", "Quasi", "Ripplefluke", "SN90", "TR82", 
                 "Upbang", "Wave", "Web","Zig")
dolphins.cluster <- rep(1, length(get.vertex.attribute(dolphins.graph, "label")))
dolphins.cluster[get.vertex.attribute(dolphins.graph, "label") %in% dolphins.truth] <- 2

pdf(file = file.path("..", "figures", "dolphins_truth.pdf"))
par(mar = c(0.5, 0.5, 0.1, 0.1))
V(dolphins.graph)$group <- dolphins.cluster
plot(dolphins.graph, 
     layout = dolphins.layout, 
     vertex.color = V(dolphins.graph)$group, 
     vertex.size = 13)
dev.off()

dolphins.df <- data.frame(t(dolphins.res$mixed.member[1,]), 
                          t(dolphins.res$mixed.member[2,]),
                          dolphins.res$clustering)
xtable(dolphins.df, digits = c(0, 4, 4, 0))

#### Examples
set.seed(314159)
sbm_prob_mat <- cbind(c(.8, .02), c(.02, .8))
sbm_graph <- sample_sbm(100, pref.matrix = sbm_prob_mat, block.sizes = c(80, 20))
sbm_adj <- as.matrix(as_adjacency_matrix(sbm_graph))
sbm.res <- cs_sbm(n_cluster = 2, net_adj = sbm_adj, burnin = 1000, size = 5000)

#### Assessments
sbm.truth <- c(rep(1, 30), rep(2, 20))
aricode::NMI(sbm.res$clustering, sbm.truth)
pdfCluster::adj.rand.index(sbm.res$clustering, sbm.truth)
